"use strict";Object.defineProperty(exports, "__esModule", {value: true}); const Scripts = {
	inherit: 'gen7',
	init() {
		for (const id in this.data.Items) {
			if (!this.data.Items[id].megaStone) continue;
			this.modData('Items', id).onTakeItem = false;
		}
	},
	canMegaEvo(pokemon) {
		if (pokemon.species.isMega || pokemon.species.isPrimal) return null;

		const item = pokemon.getItem();
		if (item.megaStone) {
			if (item.megaStone === pokemon.name) return null;
			return item.megaStone;
		} else if (pokemon.baseMoves.includes('dragonascent' )) {
			return 'Rayquaza-Mega';
		} else {
			return null;
		}
	},
	runMegaEvo(pokemon) {
		if (pokemon.species.isMega || pokemon.species.isPrimal) return false;

		const isUltraBurst = !pokemon.canMegaEvo;
		// @ts-ignore
		const species = this.getMixedSpecies(pokemon.m.originalSpecies, pokemon.canMegaEvo || pokemon.canUltraBurst);
		const side = pokemon.side;

		// Pokémon affected by Sky Drop cannot Mega Evolve. Enforce it here for now.
		for (const foeActive of side.foe.active) {
			if (foeActive.volatiles['skydrop'] && foeActive.volatiles['skydrop'].source === pokemon) {
				return false;
			}
		}

		// Do we have a proper sprite for it?
		// @ts-ignore assert non-null pokemon.canMegaEvo
		if (isUltraBurst || this.dex.getSpecies(pokemon.canMegaEvo).baseSpecies === pokemon.m.originalSpecies) {
			pokemon.formeChange(species, pokemon.getItem(), true);
		} else {
			const oSpecies = this.dex.getSpecies(pokemon.m.originalSpecies);
			// @ts-ignore
			const oMegaSpecies = this.dex.getSpecies(species.originalMega);
			pokemon.formeChange(species, pokemon.getItem(), true);
			this.add('-start', pokemon, oMegaSpecies.requiredItem || oMegaSpecies.requiredMove, '[silent]');
			if (oSpecies.types.length !== pokemon.species.types.length || oSpecies.types[1] !== pokemon.species.types[1]) {
				this.add('-start', pokemon, 'typechange', pokemon.species.types.join('/'), '[silent]');
			}
		}

		pokemon.canMegaEvo = null;
		if (isUltraBurst) pokemon.canUltraBurst = null;
		return true;
	},
	getMixedSpecies(originalSpecies, megaSpecies) {
		const oSpecies = this.dex.getSpecies(originalSpecies);
		const mSpecies = this.dex.getSpecies(megaSpecies);
		if (oSpecies.baseSpecies === mSpecies.baseSpecies) return mSpecies;
		// @ts-ignore
		const deltas = this.getMegaDeltas(mSpecies);
		// @ts-ignore
		const species = this.doGetMixedSpecies(oSpecies, deltas);
		return species;
	},
	getMegaDeltas(megaSpecies) {
		const baseSpecies = this.dex.getSpecies(megaSpecies.baseSpecies);
		const deltas








 = {
			abilities: megaSpecies.abilities,
			baseStats: {},
			weighthg: megaSpecies.weighthg - baseSpecies.weighthg,
			originalMega: megaSpecies.name,
			requiredItem: megaSpecies.requiredItem,
		};
		let stat;
		for (stat in megaSpecies.baseStats) {
			deltas.baseStats[stat] = megaSpecies.baseStats[stat] - baseSpecies.baseStats[stat];
		}
		if (megaSpecies.types.length > baseSpecies.types.length) {
			deltas.type = megaSpecies.types[1];
		} else if (megaSpecies.types.length < baseSpecies.types.length) {
			deltas.type = baseSpecies.types[0];
		} else if (megaSpecies.types[1] !== baseSpecies.types[1]) {
			deltas.type = megaSpecies.types[1];
		}
		if (megaSpecies.isMega) deltas.isMega = true;
		if (megaSpecies.isPrimal) deltas.isPrimal = true;
		return deltas;
	},
	doGetMixedSpecies(speciesOrSpeciesName, deltas) {
		if (!deltas) throw new TypeError("Must specify deltas!");
		const species = this.dex.deepClone(this.dex.getSpecies(speciesOrSpeciesName));
		species.abilities = deltas.abilities;
		if (species.types[0] === deltas.type) {
			species.types = [deltas.type];
		} else if (deltas.type) {
			species.types = [species.types[0], deltas.type];
		}
		const baseStats = species.baseStats;
		for (const statName in baseStats) {
			baseStats[statName] = this.clampIntRange(baseStats[statName] + deltas.baseStats[statName], 1, 255);
		}
		species.weighthg = Math.max(1, species.weighthg + deltas.weighthg);
		species.originalMega = deltas.originalMega;
		species.requiredItem = deltas.requiredItem;
		if (deltas.isMega) species.isMega = true;
		if (deltas.isPrimal) species.isPrimal = true;
		return species;
	},
}; exports.Scripts = Scripts;
